/* Process model C form file: sh_ip_rte_central_cpu.pr.c */
/* Portions of this file copyright 1992-2004 by OPNET Technologies, Inc. */



/* This variable carries the header into the object file */
const char sh_ip_rte_central_cpu_pr_c [] = "MIL_3_Tfile_Hdr_ 110A 30A opnet 7 44AEC64E 44AEC64E 1 luna momueti 0 0 none none 0 0 none 0 0 0 0 0 0 0 0 b56 1                                                                                                                                                                                                                                                                                                                                                                                                                ";
#include <string.h>



/* OPNET system definitions */
#include <opnet.h>



/* Header Block */

#include "ip_rte_support.h"
#include "oms_buffer_bgutil.h"
#include "oms_resource_defs.h"
#include <ip_grouping.h>

/* Transition conditions */

#define INCOMING_PACKET		(((invoke_mode == OPC_PROINV_DIRECT) && \
							  (intrpt_type == OPC_INTRPT_STRM)) || \
							 (invoke_mode == OPC_PROINV_INDIRECT))
#define OUTSTRM_TO_IP_ENCAP  0
#define OUTSTRM_TO_FIXED_HOST 1
#define OUTSTRM_TO_MOBILE_HOST 2


/* Using CPU resource management scheme */ 
#define PROCESSING_DONE		((invoke_mode == OPC_PROINV_DIRECT) && \
						 	 (intrpt_type == OPC_INTRPT_PROCESS))

#define CPU_IDLE			(cpu_is_idle == OPC_TRUE)

#define PENDING_PACKET		(oms_buffer_bgutil_num_packets_get (routing_buffer, OPC_NIL) != 0.0)

/* Macros			*/
/* Value corresponding to the symbol map 'Efficiency Mode'	*/
/* for the forwarding rate.									*/
#define IPC_EFFICIENCY_MODE_FORWARDING_RATE		-1.0

/* Global variables */

/* Flag to store whether we have written the log message	*/
/* warning the user that processing speed is set to infinity*/
static Boolean infinite_processing_log_written = OPC_FALSE;

/* Transition executives */

static void ip_rte_central_cpu_inits (void);
static void ip_rte_central_cpu_packet_arrival (void);
static void ip_rte_central_cpu_start_processing (void);
static void	ip_rte_central_cpu_send_packet (Packet* pkptr);

/* Other local routines */

static void ip_rte_cpu_error (const char * msg);
static void ip_rte_cpu_warn (const char * msg);
static void ip_rte_routing_processor_info_init (void);


/* End of Header Block */


#if !defined (VOSD_NO_FIN)
#undef	BIN
#undef	BOUT
#define	BIN		FIN_LOCAL_FIELD(_op_last_line_passed) = __LINE__ - _op_block_origin;
#define	BOUT	BIN
#define	BINIT	FIN_LOCAL_FIELD(_op_last_line_passed) = 0; _op_block_origin = __LINE__;
#else
#define	BINIT
#endif /* #if !defined (VOSD_NO_FIN) */



/* State variable definitions */
typedef struct
	{
	/* Internal state tracking for FSM */
	FSM_SYS_STATE
	/* State Variables */
	Boolean	                		cpu_is_idle                                     ;	/* OPC_TRUE when not processing a packet */
	OmsT_Buffer_Bgutil_Pool_Handle			routing_buffer_pool                             ;	/* Represents shared memory used for packet queuing	 */
	                        		                                                	/* while waiting for routing processing.			          */
	                        		                                                	/* If QoS is used, the corresponding buffers are	    */
	                        		                                                	/* also stored in that memory.						                 */
	OmsT_Buffer_Bgutil_Handle			routing_buffer                                  ;	/* Where incoming packets are stored while waiting to be processed */
	IpT_Rte_Module_Data *	  		module_data_ptr                                 ;
	Objid	                  		own_id                                          ;
	Prohandle	              		own_prohandle                                   ;
	OmsT_Resource_Handle	   		cpu_resource_handle                             ;	/* The global 'CPU' object that manages the use of CPU */
	                        		                                                	/* cycles by the various process models in the module  */
	int	                    		invoke_mode                                     ;
	double	                 		one_over_service_rate                           ;	/* Number of packets handled by routing processor per second  */
	                        		                                                	/* or number of bites per second, based on 'service_rate_pps' */
	Boolean	                		service_rate_pps                                ;	/* Where 'service_rate' is in packets or bits per second */
	double	                 		one_over_forwarding_rate                        ;	/* Forwarding/switching rate as applied to labeled datagram (MPLS) */
	                        		                                                	/* packets                                                         */
	OmsT_Resource_Job_Id	   		job_id                                          ;	/* Reference to the job being processed by the CPU module */
	Boolean	                		infinite_processing_rate                        ;	/* Flag to indicate whether the processing speed is set to infinity */
	                        		                                                	/* If this flag is set, incoming packets are forwarded immediately  */
	                        		                                                	/* bypassing the CPU. This can make the simulation faster.          */
	} sh_ip_rte_central_cpu_state;

#define pr_state_ptr            		((sh_ip_rte_central_cpu_state*) (OP_SIM_CONTEXT_PTR->_op_mod_state_ptr))
#define cpu_is_idle             		pr_state_ptr->cpu_is_idle
#define routing_buffer_pool     		pr_state_ptr->routing_buffer_pool
#define routing_buffer          		pr_state_ptr->routing_buffer
#define module_data_ptr         		pr_state_ptr->module_data_ptr
#define own_id                  		pr_state_ptr->own_id
#define own_prohandle           		pr_state_ptr->own_prohandle
#define cpu_resource_handle     		pr_state_ptr->cpu_resource_handle
#define invoke_mode             		pr_state_ptr->invoke_mode
#define one_over_service_rate   		pr_state_ptr->one_over_service_rate
#define service_rate_pps        		pr_state_ptr->service_rate_pps
#define one_over_forwarding_rate		pr_state_ptr->one_over_forwarding_rate
#define job_id                  		pr_state_ptr->job_id
#define infinite_processing_rate		pr_state_ptr->infinite_processing_rate

/* These macro definitions will define a local variable called	*/
/* "op_sv_ptr" in each function containing a FIN statement.	*/
/* This variable points to the state variable data structure,	*/
/* and can be used from a C debugger to display their values.	*/
#undef FIN_PREAMBLE_DEC
#undef FIN_PREAMBLE_CODE
#  define FIN_PREAMBLE_DEC	sh_ip_rte_central_cpu_state *op_sv_ptr;
#if defined (OPD_PARALLEL)
#  define FIN_PREAMBLE_CODE	\
		op_sv_ptr = ((sh_ip_rte_central_cpu_state *)(sim_context_ptr->_op_mod_state_ptr));
#else
#  define FIN_PREAMBLE_CODE	op_sv_ptr = pr_state_ptr;
#endif


/* Function Block */


#if !defined (VOSD_NO_FIN)
enum { _op_block_origin = __LINE__ };
#endif

/* Transition executives */

static void
ip_rte_central_cpu_inits (void)
	{
	IpT_Interface_Info *	iface_info_ptr;
	int						i;
	int						num_if;
	IpT_Group_Intf_Info*	group_info_ptr;
	int						ith_member, num_member_intfs;
	IpT_Member_Intf_Info*	member_intf_ptr;

	/** Perform appropriate initializations before reaching the wait state **/
	FIN (ip_rte_central_cpu_inits ());

	/* Obtain module memory and set up for ip_rte_int routines */
	module_data_ptr = (IpT_Rte_Module_Data *)op_pro_modmem_access ();
	ip_rte_set_procs (module_data_ptr, ip_rte_cpu_error, ip_rte_cpu_warn);

	own_id = op_id_self ();
	own_prohandle = op_pro_self ();

	/* Create pool */
	ip_rte_routing_processor_info_init ();

	/* Unless the processing speed is set to infinity, obtain	*/
	/* CPU resource handle to schedule shared processing		*/
	/* throughout the node.										*/
	if (! infinite_processing_rate)
		{
		cpu_resource_handle = Oms_Resource_Handle_Get (module_data_ptr->node_id, 
			module_data_ptr->module_id, "CPU");
		cpu_is_idle = OPC_TRUE;
		}

	/* If QoS is in place, invoke interface process model to	*/
	/* let it create its buffers using the central memory pool	*/
	/* Also take control of any stream interrupts coming on		*/
	/* each interface.											*/
	num_if = inet_rte_num_interfaces_get (module_data_ptr);
	for (i = 0; i < num_if; ++i)
		{
		iface_info_ptr = inet_rte_intf_tbl_access (module_data_ptr, i);
		
		/* Unless the stream index is invalid, register for all	*/
		/* stream interrupts on that stream. Logical interfaces	*/
		/* like loopback interfaces and tunnel interfaces		*/
		/* might have invalid stream indices.					*/

		/* In the case of interface groups, register for stream	*/
		/* interrupts on the stream indices of all member		*/
		/* interfaces.											*/
		if (ip_rte_intf_is_group (iface_info_ptr))
			{
			/* Get a handle to structure that stores group		*/
			/* related parameters.								*/
			group_info_ptr = iface_info_ptr->phys_intf_info_ptr->group_info_ptr;

			/* Get the number of member interfaces.				*/
			num_member_intfs = group_info_ptr->num_members;

			/* Loop through each member interface 				*/
			for (ith_member = 0; ith_member < num_member_intfs; ith_member++)
				{
				/* Get a handle to the ith member structure		*/
				member_intf_ptr = &(group_info_ptr->member_intf_array[ith_member]);

				/* Register for the stream interface.			*/
				op_intrpt_port_register (OPC_PORT_TYPE_STRM,
					member_intf_ptr->instrm, own_prohandle);
				}
			}
		else if (ip_rte_intf_in_port_num_get (iface_info_ptr) != IPC_PORT_NUM_INVALID)
			{
			op_intrpt_port_register (OPC_PORT_TYPE_STRM,
				ip_rte_intf_in_port_num_get (iface_info_ptr), own_prohandle);
			}
		
		if (iface_info_ptr->queuing_scheme != IpC_No_Queuing)
			{
			op_pro_invoke (iface_info_ptr->output_iface_prohandle, routing_buffer_pool);
			}
		}

	/* Take control of stream interrupts coming from the upper layers */
	op_intrpt_port_register (OPC_PORT_TYPE_STRM,
		module_data_ptr->instrm_from_ip_encap, own_prohandle);
	
	/* Check to see if there is the possibility of background utilization	*/
	/* traffic.																*/
	if ((module_data_ptr->do_bgutil) && (! infinite_processing_rate))
		{
		/* Initialize the background utilization state ptr for the central cpu*/	
		module_data_ptr->bgutil_routed_state_ptr = OPC_NIL;		
	
		/* Initialize the background utilization state ptr for keeping track */
		/* of sent and received background utilization traffic.				 */
		module_data_ptr->received_bgutil_routed_state_ptr = oms_bgutil_routed_state_create (
			!service_rate_pps, DO_NOT_SCALE);
		module_data_ptr->sent_bgutil_routed_state_ptr = oms_bgutil_routed_state_create (
			!service_rate_pps, DO_NOT_SCALE);
		
		/* Init the time at which background utilization statistics 	*/
		/* generation begins to time 0.									*/
		module_data_ptr->received_last_stat_update_time = 0.0;	
		module_data_ptr->sent_last_stat_update_time = 0.0;	
	
		/* Schedule a procedure which will be called for this process at the*/
		/* end of simulation to update background utilization statistics if	*/
		/* needed.															*/
		op_intrpt_schedule_call (OPC_INTRPT_SCHED_CALL_ENDSIM, 0, 
			ip_rte_pk_stats_update_endsim, module_data_ptr);
		}

	FOUT;
	}

static void
ip_rte_central_cpu_packet_arrival (void)
	{
	Packet *					pkptr; // = OPC_NIL;
	int							instrm;
	IpT_Rte_Ind_Ici_Fields * 	intf_ici_fdstruct_ptr ; //= OPC_NIL;
	IpT_Interface_Info *		rcvd_iface_info_ptr ; //= OPC_NIL;
	int							result;
	IpT_Dgram_Fields*			ip_dgram_fd_ptr;
	int							dest_addr;
	
	/** An incoming packet has arrived.  It might	**/
	/** be from an "upper layer", a "lower layer",	**/
	/** or generated from within ip.				**/
	FIN (ip_rte_central_cpu_packet_arrival ());
		
	if (invoke_mode == OPC_PROINV_INDIRECT)
		{
		/* Packet generated from withing IP and forwarded by our	*/
		/* parent process.											*/
		pkptr = (Packet *)op_pro_argmem_access ();
		instrm = IpC_Pk_Instrm_Child;
		}
	else
		{
		/* Packet coming from some stream */
		instrm = op_intrpt_strm ();
		pkptr = op_pk_get (instrm);
		
		if (pkptr == OPC_NIL)
			ip_rte_cpu_error ("Unable to get packet from input stream.");
	
/************************************************************************/
/*   Addition to implement the re-routing of packets to MTCP process.   */
/************************************************************************/
		if (instrm != 0)
			{
			if (instrm == 1)
				{
				op_pk_send(pkptr, OUTSTRM_TO_IP_ENCAP);
				FOUT;
				}
			else
				{
				op_pk_send(pkptr, OUTSTRM_TO_IP_ENCAP);
				FOUT;
				}
			}
/************************************************************************/
/* Addition to implement the re-routing of packets to MTCP process ends */
/************************************************************************/
	

	/* Perform standard IP processing of incoming packet 	*/
	/* 	1. Perform forwarding decision and populate the	ICI	*/
	/*	2. Populate rcvd_iface_info_ptr. It is set to NIL	*/
	/*	   if packet arrives from higher layer.				*/
	result = ip_rte_packet_arrival (module_data_ptr,
		&pkptr, instrm, &intf_ici_fdstruct_ptr, &rcvd_iface_info_ptr);
	
   	if (result == OPC_FALSE)
		{
		/* Packet was dropped in call */
		FOUT;
		}
	
	/* If the processing speed is set to infinity, forward	*/
	/* the packet immediately.								*/
	if (infinite_processing_rate)
		{
		ip_rte_central_cpu_send_packet (pkptr);
		}
	else
		{
		/* Finite processing speed.							*/

		/* Attempt to place new packet in pending queue */
		if (oms_buffer_bgutil_enqueue (routing_buffer, pkptr)
				!= OmsC_Buffer_Enqueue_Success)
			{
			/*	The insertion failed (due to a full buffer).	*/
			char intf_addr_str [IPC_ADDR_STR_LEN];

			/*  Get a printable version of the interface addr.	*/
			if (rcvd_iface_info_ptr == OPC_NIL)
				sprintf (intf_addr_str, "Higher Layer");
			else
				ip_address_print (intf_addr_str, 
					rcvd_iface_info_ptr->addr_range_ptr->address);

			/*	Issue a warning message to the sim. log.		*/
			ipnl_reswarn_pktinsert (op_pk_id (pkptr), 
				op_pk_tree_id (pkptr), intf_addr_str);

			/* Update packets dropped statistics and destroy	*/
			/* the IP datagram.									*/
			ip_rte_dgram_discard (module_data_ptr, pkptr, op_pk_ici_get (pkptr), "Buffer overflow");
			}
		}
	}
FOUT;
}





static void
ip_rte_central_cpu_start_processing (void)
	{
	Packet *	pkptr;
	Ici *		pk_ici_ptr;
	double		comp_decomp_delay = 0.0;
	double		pk_svc_time = 0.0;
	double		bgutil_packets = 0.0;
	double		bgutil_bytes = 0.0;
	double      bgutil_delay = 0.0;
	IpT_Rte_Ind_Ici_Fields * 	intf_ici_fdstruct_ptr;
	IpT_Rte_Module_Data*		iprmd_ptr;

	/** Initiate the processing of a received packet	**/
	FIN (ip_rte_central_cpu_start_processing (void));

	/* Get a handle to the packet at the top of the buffer.This	*/
	/* is the packet whose service is about to start. The		*/
	/* packet is in the central processor's buffer.				*/
	pkptr = oms_buffer_bgutil_access (routing_buffer, 0, OPC_NIL,
		&bgutil_bytes, &bgutil_packets, &bgutil_delay);
	if (pkptr == OPC_NIL)
		{
		ip_rte_cpu_error ("Unable to get packet from head of central buffer.");
		}

	
	/* Retrive the ICI associated with the packet */
	pk_ici_ptr = op_pk_ici_get (pkptr);

	/*	Get the rte_info_fields from the ici.	*/
	op_ici_attr_get (pk_ici_ptr, "rte_info_fields", &intf_ici_fdstruct_ptr);

	/*	Determine the time required to complete service of the	*/
	/*	enqueued packet, including any bgutil in front of it.	*/
	if (intf_ici_fdstruct_ptr->packet_is_labeled)
		{
		/* This is a labeled packet and the service time is 	*/
		/* based on the packet forwarding rate					*/
		pk_svc_time = one_over_forwarding_rate + bgutil_delay;
		}
	else if (service_rate_pps)
		{
   		/* rate in packets per second */
		pk_svc_time = one_over_service_rate + bgutil_delay;
		}
	else
		{
   		/* rate in bits per second */
		pk_svc_time = (one_over_service_rate * op_pk_total_size_get (pkptr)) + bgutil_delay;
		}

	/* Retrieve the decompression and/or compression delay 		*/
	/* associated with the packet. This delay will be added to 	*/
	/* the service time when scheduling the end of the service. */
	/* If the packet will not be decompressed or compressed at 	*/
	/* this processing, then the value of the delay is zero.	*/
	comp_decomp_delay = intf_ici_fdstruct_ptr->comp_decomp_delay;

	/* Reset the attribute.										*/
	if (comp_decomp_delay)
		{
		intf_ici_fdstruct_ptr->comp_decomp_delay = 0.0;
		}

	/* If the node is configured as a firewall, there can be	*/
	/* also a proxy server delay associated with the datagram.	*/
	/* Retrieve this delay, too.								*/
	if (module_data_ptr->firewall_flag == OPC_TRUE)
		{
		/* We can safely add the proxy delay to the compression	*/
		/* and decompression delay, since all of them are 		*/
		/* additional delays besides the routing delay and they	*/
		/* all will be summed up when the service completion	*/
		/* is scheduled.										*/
		comp_decomp_delay += intf_ici_fdstruct_ptr->proxy_delay;

		/* Reset the ICI field.									*/
		intf_ici_fdstruct_ptr->proxy_delay = 0.0;
		}

	/* If the node is tunneling packets, there can also	be a vpn*/
	/* tunnel delay associated with the datagram. Retrieve this */
	/* delay, too.												*/

	/* State variables cannot be passed to macros.	*/
	iprmd_ptr = module_data_ptr;
	if (ip_l2tp_vpn_is_enabled (iprmd_ptr))
		{
		/* The vpn delay can be added to the compression and    */
		/* decompression delay the same way as proxy delay.		*/
		comp_decomp_delay += intf_ici_fdstruct_ptr->vpn_delay;

		/* Reset the ICI field.									*/
		intf_ici_fdstruct_ptr->vpn_delay = 0.0;
		}
	
	/* Use the resource management package to let it decide		*/
	/* how much CPU time it takes to do the requested 			*/
	/* processing. End of processing is indicated with a		*/
	/* process interrupt.										*/
	/* Since we are dealing with one packet at a time, no need	*/
	/* to provide any structure.								*/
	job_id = Oms_Resource_Job_Insert (cpu_resource_handle, own_prohandle, 
		OPC_NIL, pk_svc_time + comp_decomp_delay);

	if (Oms_Resource_Job_Id_Equal (job_id, OMSC_JOB_ID_INVALID))
		{
		ip_rte_cpu_warn ("Unable to schedule service completion.");
		}
	else
		{
		/*	Mark the server to be busy now.						*/
		cpu_is_idle = OPC_FALSE;
    	}

	FOUT;
	}


static void
ip_rte_central_cpu_send_packet (Packet* pkptr)
	{
	OpT_Packet_Size			packet_size;
	Ici 					*pk_ici_ptr;
	IpT_Dgram_Fields*		pk_fd_ptr;
	IpT_Rte_Ind_Ici_Fields  *intf_ici_fdstruct_ptr;

	/** Send processed packet to appropriate level	**/
	FIN (ip_rte_central_cpu_send_packet (pkptr));

	/* Unless the processing speed is set to infinity,	*/
	/* we need to extract the packet from the buffer.	*/

	if (! infinite_processing_rate)
		{
		/* Destroy job ID. Resource package does not 		*/
		/* destroy the memory allocated to store this info.	*/
		/* if there is no ici expected , the job id is      */
		/* destroyed by resource package					*/
		/* op_prg_mem_free (job_id); 					    */
		
		/* Block outgoing background traffic during the service of this packet. */
		oms_buffer_bgutil_outgoing_flow_control(routing_buffer, OPC_FALSE);
		
		/* Extract packet at head of queue - this is the packet just	*/
		/* finishing service. 											*/
		pkptr = oms_buffer_bgutil_dequeue (routing_buffer);

		/* Reset the outgoing background traffic flag. */
		oms_buffer_bgutil_outgoing_flow_control (routing_buffer, OPC_TRUE);
		
		if (pkptr == OPC_NIL)
			ip_rte_cpu_error ("Unable to get packet from head of central buffer.");
		
		/* The central cpu becomes available again.	*/
		cpu_is_idle = OPC_TRUE;
		}
	
	/* Retrieve the ICI associated with the packet.	*/
	pk_ici_ptr = op_pk_ici_get (pkptr);
	
	/*	Get the data structure stored in the "rte_info_fields" field.	*/
	if (op_ici_attr_exists (pk_ici_ptr, "rte_info_fields") == OPC_TRUE)
		{
		op_ici_attr_get (pk_ici_ptr, "rte_info_fields", &intf_ici_fdstruct_ptr);
		}
	else
		{
		ip_rte_cpu_error ("Unable to obtain routing information from packet.");
		}
	
	/* Get a handle to the data structure stored in "fields" field. */
	/* This structure contains all the protocol header information 	*/
	/* as well as some internal information attached to each IP 	*/
	/* datagram for simulation efficiency. 			 				*/
	op_pk_nfd_access (pkptr, "fields", &pk_fd_ptr);
	
	/* Decompress the packet if necessary.	*/
	ip_rte_decompress (pkptr, intf_ici_fdstruct_ptr, pk_fd_ptr);

	/* Update the packet size. */
	packet_size = op_pk_total_size_get (pkptr);

	/* Compress datagram if specified */
	module_data_ptr->dgram_compressed_size = ip_rte_compress (
		module_data_ptr, pkptr, packet_size, intf_ici_fdstruct_ptr, pk_fd_ptr);

	/* Call the function that would forward the packet appropriately*/
	ip_rte_packet_send (module_data_ptr, pkptr, pk_ici_ptr,
		intf_ici_fdstruct_ptr, IpC_Rte_Process_Type_Central_Cpu, OPC_NIL);

	FOUT;
	}


/* Other local routines */

static void
ip_rte_cpu_error (const char *msg)
	{
	/** Print an error message and exit the simulation. **/
	FIN (ip_rte_cpu_error (msg));

	op_sim_end (
		"Error in IP central forwarding process model (ip_rte_central_cpu):",
		msg,
		"Check simulation log messages from the IP model for more information.",
		"Simulation Logging can be enabled/disabled in the Simulation Editor."); 
	FOUT;
	}

static void
ip_rte_cpu_warn (const char *msg)
	{
	/** Print a warning message and resume. **/
	FIN (ip_rte_cpu_warn (msg));

	op_sim_message ("Warning from IP central forwarding process model (ip_rte_central_cpu):", msg);

	FOUT;
	}


static void
ip_rte_routing_processor_info_init (void)
	{
	Objid				ip_proc_info_objid;
	Objid				compound_objid;
	Compcode			return_status;
	double				buffer_size;
	OmsT_Buffer_Create	status;
	OmsT_Buffer_Stats	stat_array;
	int					i;
	double				service_rate;
	double				forwarding_rate;

	const char *	stat_names [OmsC_Buffer_Stat_Number] =
		{
		"IP Processor.Forwarding Memory Overflows",
		"IP Processor.Forwarding Memory Free Size (bytes)",
		"", /* ignored */
		"IP Processor.Forwarding Memory Queue Size (bytes)",
		"IP Processor.Forwarding Memory Queue Size (packets)",
		"IP Processor.Forwarding Memory Queue Size (incoming bytes)",
		"IP Processor.Forwarding Memory Queue Size (incoming packets)",
		"IP Processor.Forwarding Memory Queuing Delay",
		"", /* delay variation ignored */			
		};

	/** Extract routing processor information and perform appropriate initializtions **/
	FIN (ip_rte_routing_processor_info_init ());

	return_status = op_ima_obj_attr_get (own_id, 
		"ip processing information", &ip_proc_info_objid);
	if (return_status == OPC_COMPCODE_FAILURE)
		ip_rte_cpu_error ("Unable to get forwarding information from attribute.");

	compound_objid = op_topo_child (ip_proc_info_objid, OPC_OBJTYPE_GENERIC, 0);

	/*	Get assigned value of server processing rate.				*/
	return_status = op_ima_obj_attr_get (compound_objid, 
		"Datagram Forwarding Rate", &service_rate);
   	if (return_status == OPC_COMPCODE_FAILURE)
		ip_rte_cpu_error ("Unable to get server datagram forwarding rate from attribute.");

	/* If the service rate is set to infinity, we don't need	*/
	/* to create a buffer to hold the packets while they are	*/
	/* being serviced.											*/
	if (IPC_EFFICIENCY_MODE_FORWARDING_RATE == service_rate)
		{
		/* Set the flag indicating that the processing speed is	*/
		/* infinite.											*/
		infinite_processing_rate = OPC_TRUE;

		/* Write a log message warning the user that CPU stats	*/
		/* will not be recorded because the processing speed	*/
		/* is set to infinity.									*/
		if (OPC_FALSE == infinite_processing_log_written)
			{
			/* The log message has not been written already.	*/
			ipnl_cpu_stats_not_written_log_write ();

			/* Do not write the log message again.				*/
			infinite_processing_log_written = OPC_TRUE;
			}

		FOUT;
		}

	/* Finite processing speed.									*/
	infinite_processing_rate = OPC_FALSE;

	/* Store 1/service rate to avoid doing division each time */
	one_over_service_rate = 1.0 / service_rate;
    module_data_ptr->service_rate = service_rate;
	
	/* MPLS forwarding rate									*/
	return_status = op_ima_obj_attr_get (compound_objid,
			"Datagram Switching Rate", &forwarding_rate);
	if (return_status == OPC_COMPCODE_FAILURE)
		ip_rte_cpu_error ("Unable to get Datagram Switching rate from the attribute.");
	one_over_forwarding_rate = 1.0/forwarding_rate;

	return_status = op_ima_obj_attr_get (compound_objid,
		"Forwarding Rate Units", &service_rate_pps);
	if (return_status == OPC_COMPCODE_FAILURE)
		ip_rte_cpu_error ("Unable to get units of forwarding rate from attribute.");

	/* Find out routing processing buffer size and create corresponding buffer	*/
	return_status = op_ima_obj_attr_get (compound_objid,
		"Memory Size", &buffer_size);
	if (return_status == OPC_COMPCODE_FAILURE)
		ip_rte_cpu_error ("Unable to get forwarding memory size from attribute.");

	/* The buffer pool will be used for the routing queue, as well	*/
	/* as the QoS buffers, if no slots are used.					*/
	/* The pool size is based on the routing processor memory size	*/
	/* The pool does not have any size limitations.	*/
	/* The various buffers will have their own.		*/
	routing_buffer_pool = oms_buffer_bgutil_pool_create (
		buffer_size,			/* number of bytes */
		OmsC_Buffer_Infinite,	/* number of packets */
		OPC_NIL, "Routing Processor Pool", OPC_NIL, &status);
	if (routing_buffer_pool == OPC_NIL)
		{
		/* Something went wrong, report and terminate */
		switch (status)
			{
			case OmsC_Buffer_Create_Invalid_Byte_Size:
				op_sim_end (
					"Unable to create IP shared memory pool due to an invalid byte size", 
					"", "", "");
				break;
			case OmsC_Buffer_Create_Invalid_Packet_Size:
				op_sim_end (
					"Unable to create IP shared memory pool due to an invalid packet size", 
					"", "", "");
				break;
			case OmsC_Buffer_Create_Memory_Alloc_Failed:
				op_sim_end (
					"Unable to create IP shared memory pool due to an internal memory allocation failure", 
					"", "", "");
				break;
			}
		}

	/* All stats but free size in packets (since no upper limit) */
	for (i = 0; i < OmsC_Buffer_Stat_Number; i++)
		{
		if (stat_names [i][0] == '\0')
    		stat_array [i].used = OPC_FALSE;
		else
			{
    		stat_array [i].used = OPC_TRUE;
	    	stat_array [i].shandle = op_stat_reg (stat_names [i], 
				OPC_STAT_INDEX_NONE, OPC_STAT_LOCAL);
			}
		}
	

	routing_buffer = oms_buffer_bgutil_create (routing_buffer_pool,
		OmsC_Buffer_Parent_Limit, OmsC_Buffer_Parent_Limit,
		module_data_ptr->do_bgutil, 
		service_rate, service_rate_pps,
		&stat_array, "Routing Processor Buffer", OPC_NIL, /* No user data */
		&status);
	if (routing_buffer == OPC_NIL)
		ip_rte_cpu_error ("Unable to create routing processor memory buffer.");

	FOUT;
	}

/* End of Function Block */

/* Undefine optional tracing in FIN/FOUT/FRET */
/* The FSM has its own tracing code and the other */
/* functions should not have any tracing.		  */
#undef FIN_TRACING
#define FIN_TRACING

#undef FOUTRET_TRACING
#define FOUTRET_TRACING

#if defined (__cplusplus)
extern "C" {
#endif
	void sh_ip_rte_central_cpu (OP_SIM_CONTEXT_ARG_OPT);
	VosT_Obtype _op_sh_ip_rte_central_cpu_init (int * init_block_ptr);
	VosT_Address _op_sh_ip_rte_central_cpu_alloc (VOS_THREAD_INDEX_ARG_COMMA VosT_Obtype, int);
	void _op_sh_ip_rte_central_cpu_diag (OP_SIM_CONTEXT_ARG_OPT);
	void _op_sh_ip_rte_central_cpu_terminate (OP_SIM_CONTEXT_ARG_OPT);
	void _op_sh_ip_rte_central_cpu_svar (void *, const char *, void **);


	VosT_Obtype Vos_Define_Object_Prstate (const char * _op_name, unsigned int _op_size);
	VosT_Address Vos_Alloc_Object_MT (VOS_THREAD_INDEX_ARG_COMMA VosT_Obtype _op_ob_hndl);
	VosT_Fun_Status Vos_Poolmem_Dealloc_MT (VOS_THREAD_INDEX_ARG_COMMA VosT_Address _op_ob_ptr);
#if defined (__cplusplus)
} /* end of 'extern "C"' */
#endif




/* Process model interrupt handling procedure */


void
sh_ip_rte_central_cpu (OP_SIM_CONTEXT_ARG_OPT)
	{

#if !defined (VOSD_NO_FIN)
	int _op_block_origin = 0;
#endif
	FIN_MT (sh_ip_rte_central_cpu ());
		{
		/* Temporary Variables */
		/* For transition conditions */
		int			intrpt_type;
		Prohandle	invoke_prohandle;
		/* End of Temporary Variables */


		FSM_ENTER ("sh_ip_rte_central_cpu")

		FSM_BLOCK_SWITCH
			{
			/*---------------------------------------------------------*/
			/** state (init) enter executives **/
			FSM_STATE_ENTER_FORCED_NOLABEL (0, "init", "sh_ip_rte_central_cpu [init enter execs]")

			/** state (init) exit executives **/
			FSM_STATE_EXIT_FORCED (0, "init", "sh_ip_rte_central_cpu [init exit execs]")


			/** state (init) transition processing **/
			FSM_TRANSIT_FORCE (1, state1_enter_exec, ip_rte_central_cpu_inits ();, "default", "ip_rte_central_cpu_inits ()", "init", "ip_central_cpu_wait", "sh_ip_rte_central_cpu [init -> ip_central_cpu_wait : default / ip_rte_central_cpu_inits ()]")
				/*---------------------------------------------------------*/



			/** state (ip_central_cpu_wait) enter executives **/
			FSM_STATE_ENTER_UNFORCED (1, "ip_central_cpu_wait", state1_enter_exec, "sh_ip_rte_central_cpu [ip_central_cpu_wait enter execs]")

			/** blocking after enter executives of unforced state. **/
			FSM_EXIT (3,"sh_ip_rte_central_cpu")


			/** state (ip_central_cpu_wait) exit executives **/
			FSM_STATE_EXIT_UNFORCED (1, "ip_central_cpu_wait", "sh_ip_rte_central_cpu [ip_central_cpu_wait exit execs]")
				FSM_PROFILE_SECTION_IN ("sh_ip_rte_central_cpu [ip_central_cpu_wait exit execs]", state1_exit_exec)
				{
				/* determine the interuppt type */
				intrpt_type = op_intrpt_type ();
				
				/* Determine if we are being invoked by one of the child proceses	*/
				/* maintained by IP. Note that "ip_icmp" is used for ICMP messages	*/
				/* (currently only supports "ping") and "oms_basetraf_src" is used	*/
				/* background utilization traffic specification generation.			*/
				invoke_prohandle = op_pro_invoker (own_prohandle, &invoke_mode);
				if ((invoke_mode != OPC_PROINV_INDIRECT) && (invoke_mode != OPC_PROINV_DIRECT))
					{
					ip_rte_cpu_error ("Unable to determine how IP process got invoked."); 	
					}
				}
				FSM_PROFILE_SECTION_OUT (state1_exit_exec)


			/** state (ip_central_cpu_wait) transition processing **/
			FSM_PROFILE_SECTION_IN ("sh_ip_rte_central_cpu [ip_central_cpu_wait trans conditions]", state1_trans_conds)
			FSM_INIT_COND (PROCESSING_DONE)
			FSM_TEST_COND ((!infinite_processing_rate) && INCOMING_PACKET)
			FSM_TEST_COND (infinite_processing_rate && INCOMING_PACKET)
			FSM_TEST_LOGIC ("ip_central_cpu_wait")
			FSM_PROFILE_SECTION_OUT (state1_trans_conds)

			FSM_TRANSIT_SWITCH
				{
				FSM_CASE_TRANSIT (0, 4, state4_enter_exec, ip_rte_central_cpu_send_packet (OPC_NIL);, "PROCESSING_DONE", "ip_rte_central_cpu_send_packet (OPC_NIL)", "ip_central_cpu_wait", "check_next", "sh_ip_rte_central_cpu [ip_central_cpu_wait -> check_next : PROCESSING_DONE / ip_rte_central_cpu_send_packet (OPC_NIL)]")
				FSM_CASE_TRANSIT (1, 2, state2_enter_exec, ip_rte_central_cpu_packet_arrival ();, "(!infinite_processing_rate) && INCOMING_PACKET", "ip_rte_central_cpu_packet_arrival ()", "ip_central_cpu_wait", "check_work", "sh_ip_rte_central_cpu [ip_central_cpu_wait -> check_work : (!infinite_processing_rate) && INCOMING_PACKET / ip_rte_central_cpu_packet_arrival ()]")
				FSM_CASE_TRANSIT (2, 1, state1_enter_exec, ip_rte_central_cpu_packet_arrival ();, "infinite_processing_rate && INCOMING_PACKET", "ip_rte_central_cpu_packet_arrival ()", "ip_central_cpu_wait", "ip_central_cpu_wait", "sh_ip_rte_central_cpu [ip_central_cpu_wait -> ip_central_cpu_wait : infinite_processing_rate && INCOMING_PACKET / ip_rte_central_cpu_packet_arrival ()]")
				}
				/*---------------------------------------------------------*/



			/** state (check_work) enter executives **/
			FSM_STATE_ENTER_FORCED (2, "check_work", state2_enter_exec, "sh_ip_rte_central_cpu [check_work enter execs]")

			/** state (check_work) exit executives **/
			FSM_STATE_EXIT_FORCED (2, "check_work", "sh_ip_rte_central_cpu [check_work exit execs]")


			/** state (check_work) transition processing **/
			FSM_PROFILE_SECTION_IN ("sh_ip_rte_central_cpu [check_work trans conditions]", state2_trans_conds)
			FSM_INIT_COND (CPU_IDLE && PENDING_PACKET)
			FSM_DFLT_COND
			FSM_TEST_LOGIC ("check_work")
			FSM_PROFILE_SECTION_OUT (state2_trans_conds)

			FSM_TRANSIT_SWITCH
				{
				FSM_CASE_TRANSIT (0, 3, state3_enter_exec, ;, "CPU_IDLE && PENDING_PACKET", "", "check_work", "start_proc", "sh_ip_rte_central_cpu [check_work -> start_proc : CPU_IDLE && PENDING_PACKET / ]")
				FSM_CASE_TRANSIT (1, 1, state1_enter_exec, ;, "default", "", "check_work", "ip_central_cpu_wait", "sh_ip_rte_central_cpu [check_work -> ip_central_cpu_wait : default / ]")
				}
				/*---------------------------------------------------------*/



			/** state (start_proc) enter executives **/
			FSM_STATE_ENTER_FORCED (3, "start_proc", state3_enter_exec, "sh_ip_rte_central_cpu [start_proc enter execs]")

			/** state (start_proc) exit executives **/
			FSM_STATE_EXIT_FORCED (3, "start_proc", "sh_ip_rte_central_cpu [start_proc exit execs]")


			/** state (start_proc) transition processing **/
			FSM_TRANSIT_FORCE (1, state1_enter_exec, ip_rte_central_cpu_start_processing ();, "default", "ip_rte_central_cpu_start_processing ()", "start_proc", "ip_central_cpu_wait", "sh_ip_rte_central_cpu [start_proc -> ip_central_cpu_wait : default / ip_rte_central_cpu_start_processing ()]")
				/*---------------------------------------------------------*/



			/** state (check_next) enter executives **/
			FSM_STATE_ENTER_FORCED (4, "check_next", state4_enter_exec, "sh_ip_rte_central_cpu [check_next enter execs]")

			/** state (check_next) exit executives **/
			FSM_STATE_EXIT_FORCED (4, "check_next", "sh_ip_rte_central_cpu [check_next exit execs]")


			/** state (check_next) transition processing **/
			FSM_PROFILE_SECTION_IN ("sh_ip_rte_central_cpu [check_next trans conditions]", state4_trans_conds)
			FSM_INIT_COND (PENDING_PACKET)
			FSM_DFLT_COND
			FSM_TEST_LOGIC ("check_next")
			FSM_PROFILE_SECTION_OUT (state4_trans_conds)

			FSM_TRANSIT_SWITCH
				{
				FSM_CASE_TRANSIT (0, 3, state3_enter_exec, ;, "PENDING_PACKET", "", "check_next", "start_proc", "sh_ip_rte_central_cpu [check_next -> start_proc : PENDING_PACKET / ]")
				FSM_CASE_TRANSIT (1, 1, state1_enter_exec, ;, "default", "", "check_next", "ip_central_cpu_wait", "sh_ip_rte_central_cpu [check_next -> ip_central_cpu_wait : default / ]")
				}
				/*---------------------------------------------------------*/



			}


		FSM_EXIT (0,"sh_ip_rte_central_cpu")
		}
	}




void
_op_sh_ip_rte_central_cpu_diag (OP_SIM_CONTEXT_ARG_OPT)
	{
	/* No Diagnostic Block */
	}




void
_op_sh_ip_rte_central_cpu_terminate (OP_SIM_CONTEXT_ARG_OPT)
	{

#if !defined (VOSD_NO_FIN)
	int _op_block_origin = __LINE__;
#endif

	FIN_MT (_op_sh_ip_rte_central_cpu_terminate ())


	/* No Termination Block */

	Vos_Poolmem_Dealloc_MT (OP_SIM_CONTEXT_THREAD_INDEX_COMMA pr_state_ptr);

	FOUT
	}


/* Undefine shortcuts to state variables to avoid */
/* syntax error in direct access to fields of */
/* local variable prs_ptr in _op_sh_ip_rte_central_cpu_svar function. */
#undef cpu_is_idle
#undef routing_buffer_pool
#undef routing_buffer
#undef module_data_ptr
#undef own_id
#undef own_prohandle
#undef cpu_resource_handle
#undef invoke_mode
#undef one_over_service_rate
#undef service_rate_pps
#undef one_over_forwarding_rate
#undef job_id
#undef infinite_processing_rate

#undef FIN_PREAMBLE_DEC
#undef FIN_PREAMBLE_CODE

#define FIN_PREAMBLE_DEC
#define FIN_PREAMBLE_CODE

VosT_Obtype
_op_sh_ip_rte_central_cpu_init (int * init_block_ptr)
	{

#if !defined (VOSD_NO_FIN)
	int _op_block_origin = 0;
#endif
	VosT_Obtype obtype = OPC_NIL;
	FIN_MT (_op_sh_ip_rte_central_cpu_init (init_block_ptr))

	obtype = Vos_Define_Object_Prstate ("proc state vars (sh_ip_rte_central_cpu)",
		sizeof (sh_ip_rte_central_cpu_state));
	*init_block_ptr = 0;

	FRET (obtype)
	}

VosT_Address
_op_sh_ip_rte_central_cpu_alloc (VOS_THREAD_INDEX_ARG_COMMA VosT_Obtype obtype, int init_block)
	{

#if !defined (VOSD_NO_FIN)
	int _op_block_origin = 0;
#endif
	sh_ip_rte_central_cpu_state * ptr;
	FIN_MT (_op_sh_ip_rte_central_cpu_alloc (obtype))

	ptr = (sh_ip_rte_central_cpu_state *)Vos_Alloc_Object_MT (VOS_THREAD_INDEX_COMMA obtype);
	if (ptr != OPC_NIL)
		{
		ptr->_op_current_block = init_block;
#if defined (OPD_ALLOW_ODB)
		ptr->_op_current_state = "sh_ip_rte_central_cpu [init enter execs]";
#endif
		}
	FRET ((VosT_Address)ptr)
	}



void
_op_sh_ip_rte_central_cpu_svar (void * gen_ptr, const char * var_name, void ** var_p_ptr)
	{
	sh_ip_rte_central_cpu_state		*prs_ptr;

	FIN_MT (_op_sh_ip_rte_central_cpu_svar (gen_ptr, var_name, var_p_ptr))

	if (var_name == OPC_NIL)
		{
		*var_p_ptr = (void *)OPC_NIL;
		FOUT
		}
	prs_ptr = (sh_ip_rte_central_cpu_state *)gen_ptr;

	if (strcmp ("cpu_is_idle" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->cpu_is_idle);
		FOUT
		}
	if (strcmp ("routing_buffer_pool" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->routing_buffer_pool);
		FOUT
		}
	if (strcmp ("routing_buffer" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->routing_buffer);
		FOUT
		}
	if (strcmp ("module_data_ptr" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->module_data_ptr);
		FOUT
		}
	if (strcmp ("own_id" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->own_id);
		FOUT
		}
	if (strcmp ("own_prohandle" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->own_prohandle);
		FOUT
		}
	if (strcmp ("cpu_resource_handle" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->cpu_resource_handle);
		FOUT
		}
	if (strcmp ("invoke_mode" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->invoke_mode);
		FOUT
		}
	if (strcmp ("one_over_service_rate" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->one_over_service_rate);
		FOUT
		}
	if (strcmp ("service_rate_pps" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->service_rate_pps);
		FOUT
		}
	if (strcmp ("one_over_forwarding_rate" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->one_over_forwarding_rate);
		FOUT
		}
	if (strcmp ("job_id" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->job_id);
		FOUT
		}
	if (strcmp ("infinite_processing_rate" , var_name) == 0)
		{
		*var_p_ptr = (void *) (&prs_ptr->infinite_processing_rate);
		FOUT
		}
	*var_p_ptr = (void *)OPC_NIL;

	FOUT
	}

